/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/
package org.jboss.beans.metadata.plugins;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.jboss.beans.metadata.spi.BeanMetaData;
import org.jboss.beans.metadata.spi.ClassLoaderMetaData;
import org.jboss.beans.metadata.spi.ConstructorMetaData;
import org.jboss.beans.metadata.spi.LifecycleMetaData;
import org.jboss.beans.metadata.spi.PropertyMetaData;
import org.jboss.dependency.spi.ControllerMode;
import org.jboss.dependency.spi.ControllerState;
import org.jboss.util.JBossObject;
import org.jboss.util.JBossStringBuilder;

/**
 * Metadata for a bean.
 * 
 * @author <a href="adrian@jboss.com">Adrian Brock</a>
 * @version $Revision: 1.18 $
 */
public class AbstractBeanMetaData extends AbstractFeatureMetaData implements BeanMetaData
{
   /** The bean class name */
   protected String bean;

   /** The name of this instance */
   protected String name;
   
   /** The mode */
   protected ControllerMode mode = null;
   
   /** The properties configuration Set<PropertyMetaData> */
   private Set properties;

   /** The bean ClassLoader */
   protected ClassLoaderMetaData classLoader;

   /** The constructor */
   protected ConstructorMetaData constructor;
   
   /** The create lifecycle */
   protected LifecycleMetaData create;
   
   /** The start lifecycle */
   protected LifecycleMetaData start;
   
   /** The stop lifecycle */
   protected LifecycleMetaData stop;
   
   /** The destroy lifecycle */
   protected LifecycleMetaData destroy;
   
   /** What the bean demands Set<DemandMetaData> */
   protected Set demands;
   
   /** What the bean supplies Set<SupplyMetaData> */
   protected Set supplies;

   protected Map metaData;
   /**
    * Create a new bean meta data
    */
   public AbstractBeanMetaData()
   {
      super();
   }

   /**
    * Create a new bean meta data
    * 
    * @param bean the bean class name
    */
   public AbstractBeanMetaData(String bean)
   {
      this.bean = bean;
   }
   /**
    * Create a new bean meta data
    * 
    * @param name the name
    * @param bean the bean class name
    */
   public AbstractBeanMetaData(String name, String bean)
   {
      this.name = name;
      this.bean = bean;
   }

   /**
    * Set the bean.
    * 
    * @param bean The bean to set.
    */
   public void setBean(String bean)
   {
      this.bean = bean;
      flushJBossObjectCache();
   }

   /**
    * Get a property
    * 
    * @param name the name
    * @return the property name
    */
   public PropertyMetaData getProperty(String name)
   {
      if (properties != null && properties.size() > 0)
      {
         for (Iterator i = properties.iterator(); i.hasNext();)
         {
            AbstractPropertyMetaData prop = (AbstractPropertyMetaData) i.next();
            if ("properties".equals(prop.getName()))
            {
               return prop;
            }
         }
      }
      return null;
   }

   /**
    * Add a property
    * 
    * @param property the property
    */
   public void addProperty(PropertyMetaData property)
   {
      if (property == null)
         throw new IllegalArgumentException("Null property");
      if (properties == null)
         properties = new HashSet();
      properties.add(property);
      flushJBossObjectCache();
   }
   
   /**
    * Set the propertiess.
    * 
    * @param properties Set<PropertiesMetaData>
    */
   public void setProperties(Set properties)
   {
      this.properties = properties;
      flushJBossObjectCache();
   }

   public ClassLoaderMetaData getClassLoader()
   {
      return classLoader;
   }

   public void setClassLoader(ClassLoaderMetaData classLoader)
   {
      this.classLoader = classLoader;
   }

   /**
    * Set the constructor
    * 
    * @param constructor the constructor metadata
    */
   public void setConstructor(ConstructorMetaData constructor)
   {
      this.constructor = constructor;
   }

   /**
    * Set what the bean demands.
    * 
    * @param demands Set<DemandMetaData>
    */
   public void setDemands(Set demands)
   {
      this.demands = demands;
      flushJBossObjectCache();
   }
   
   /**
    * Set what the bean supplies.
    * 
    * @param supplies Set<SupplyMetaData>
    */
   public void setSupplies(Set supplies)
   {
      this.supplies = supplies;
      flushJBossObjectCache();
   }
   
   public String getBean()
   {
      return bean;
   }

   public String getName()
   {
      return name;
   }

   /**
    * Set the name.
    * 
    * @param name The name to set.
    */
   public void setName(String name)
   {
      this.name = name;
      flushJBossObjectCache();
   }

   public ControllerMode getMode()
   {
      return mode;
   }
   
   public void setMode(ControllerMode mode)
   {
      this.mode = mode;
      flushJBossObjectCache();
   }
   
   public Set getProperties()
   {
      return properties;
   }
   
   public ConstructorMetaData getConstructor()
   {
      return constructor;
   }
   
   public LifecycleMetaData getCreate()
   {
      return create;
   }
   
   /**
    * Set the lifecycle metadata
    * 
    * @param lifecycle the lifecycle metadata
    */
   public void setCreate(LifecycleMetaData lifecycle)
   {
      lifecycle.setState(ControllerState.CREATE);
      this.create = lifecycle;
   }
   
   public LifecycleMetaData getStart()
   {
      return start;
   }
   
   /**
    * Set the start metadata
    * 
    * @param lifecycle the lifecycle metadata
    */
   public void setStart(LifecycleMetaData lifecycle)
   {
      lifecycle.setState(ControllerState.START);
      this.start = lifecycle;
   }
   
   public LifecycleMetaData getStop()
   {
      return stop;
   }
   
   /**
    * Set the stop metadata
    * 
    * @param lifecycle the lifecycle metadata
    */
   public void setStop(LifecycleMetaData lifecycle)
   {
      lifecycle.setState(ControllerState.START);
      this.stop = lifecycle;
   }
   
   public LifecycleMetaData getDestroy()
   {
      return destroy;
   }
   
   /**
    * Set the destroy metadata
    * 
    * @param lifecycle the lifecycle metadata
    */
   public void setDestroy(LifecycleMetaData lifecycle)
   {
      lifecycle.setState(ControllerState.CREATE);
      this.destroy = lifecycle;
   }
   
   public Set getDemands()
   {
      return demands;
   }
   
   public Set getSupplies()
   {
      return supplies;
   }
   
   public Iterator getChildren()
   {
      ArrayList list = new ArrayList();
      getChildren(list);
      return list.iterator();
   }
   
   /**
    * Get the children
    * 
    * @param collection the children
    */
   protected void getChildren(Collection collection)
   {
      if (classLoader != null)
         collection.add(classLoader);
      if (constructor != null)
         collection.add(constructor);
      if (properties != null)
         collection.addAll(properties);
      if (create != null)
         collection.add(create);
      if (start != null)
         collection.add(start);
      if (stop != null)
         collection.add(stop);
      if (destroy != null)
         collection.add(destroy);
      if (demands != null)
         collection.addAll(demands);
      if (supplies != null)
         collection.addAll(supplies);
      if (metaData != null)
         collection.add(metaData);
   }
   
   public void toString(JBossStringBuilder buffer)
   {
      buffer.append("name=").append(name);
      buffer.append(" bean=").append(bean);
      buffer.append(" properties=");
      JBossObject.list(buffer, properties);
      if (classLoader != null)
         buffer.append(" classLoader=").append(classLoader);
      buffer.append(" constructor=").append(constructor);
      if (create != null)
         buffer.append(" create=").append(create);
      if (start != null)
         buffer.append(" start=").append(start);
      if (stop != null)
         buffer.append(" stop=").append(stop);
      if (destroy != null)
         buffer.append(" destroy=").append(destroy);
      if (demands != null)
      {
         buffer.append(" demands=");
         JBossObject.list(buffer, demands);
      }
      super.toString(buffer);
      if (supplies != null)
      {
         buffer.append(" supplies=");
         JBossObject.list(buffer, supplies);
      }
   }
   
   public void toShortString(JBossStringBuilder buffer)
   {
      buffer.append(bean);
      buffer.append('/');
      buffer.append(name);
   }
}
